{% extends "security/base.html" %}
{% from "security/_macros.html" import render_field_with_errors, render_field, header %}

{% block content %}
{% include "security/_messages.html" %}
{{ header(_('Register')) }}
<form action="{{ url_for_security('register') }}" method="POST" name="register_user_form">
  {{ register_user_form.hidden_tag() }}
  {{ render_field_with_errors(register_user_form.email) }}
  {{ render_field_with_errors(register_user_form.username) }}
  {{ render_field_with_errors(register_user_form.password) }}
  {% if register_user_form.password_confirm %}
    {{ render_field_with_errors(register_user_form.password_confirm) }}
  {% endif %}
  {{ render_field(register_user_form.submit, class_="button is-link") }}
</form>
{% endblock %}
